﻿using MathNet.Numerics.LinearAlgebra;
using Microsoft.ML;
using PLplot;
using System;
using System.Collections.Generic;
using System.Linq;

class Program
{
    static void Main(string[] args)
    {
        MLContext ctx = new MLContext();

        IDataView data = ctx.Data.LoadFromTextFile<ClickData>("click.csv", separatorChar: ',', hasHeader: true);

        var xyList = ctx.Data.CreateEnumerable<ClickData>(data, false);

        double[] xData = xyList.Select(xy => (double)xy.X).ToArray();
        double[] yData = xyList.Select(xy => (double)xy.Y).ToArray();

        Func<double, double>[] drawFunc = new Func<double, double>[3];

        {
            // y = theta0 + (theta1 * x)
            double[] theta = GetPolynomial(xData, yData, 1);
            Console.WriteLine($"[method1] y = {theta[1]} + {theta[0]} * x");

            Func<double, double> func =
                (x) => theta[1] + theta[0] * x;

            drawFunc[0] = func;
            ReportError(xData, yData, func);
        }


        {
            // y = theta0 + (theta1 * x) + (theta2 * x * x)
            double[] theta = GetPolynomial(xData, yData, 2);
            Console.WriteLine($"[method2] y = {theta[2]} + {theta[1]} * x + {theta[0]} * x^2");

            Func<double, double> func =
                (x) => theta[2] + theta[1] * x + theta[0] * x * x;

            drawFunc[1] = func;
            ReportError(xData, yData, func);
        }


        {
            // y = theta0 + (theta1 * x) + (theta2 * x * x) + (theta3 * x * x * x)
            double[] theta = GetPolynomial(xData, yData, 3);
            Console.WriteLine($"[method3] y = {theta[2]} + {theta[1]} * x + {theta[0]} * x^2 + {theta[0]} * x^3");

            Func<double, double> func =
                (x) => theta[3] + theta[2] * x + theta[1] * x * x + theta[0] * x * x * x;

            drawFunc[2] = func;
            ReportError(xData, yData, func);
        }

        DrawPlotChart(xData, yData, drawFunc);
    }

    private static void ReportError(double[] xData, double[] yData, Func<double, double> func)
    {
        double error = 0.0;

        for (int i = 0; i < xData.Length; i ++)
        {
            double diff = yData[i] - func(xData[i]);
            error += (diff * diff);
        }

        Console.WriteLine("Error: " + error);
    }

    private static double[] GetPolynomial(double[] xData, double[] yData, int numberOfEfficient)
    {
        Matrix<double> matA = CreateMatrix.DenseOfColumnMajor(xData.Count(), 1, xData);
        Vector<double> add1 = Vector<double>.Build.DenseOfArray(Enumerable.Repeat(1.0, xData.Count()).ToArray());
        matA = matA.InsertColumn(1, add1);

        for (int i = 1; i < numberOfEfficient; i++)
        {
            double[] newColumnData = xData.Select((elem) => Math.Pow(elem, i + 1)).ToArray();
            Vector<double> addX = Vector<double>.Build.DenseOfArray(newColumnData);
            matA = matA.InsertColumn(0, addX);
        }

        Console.WriteLine(matA);
        Matrix<double> matB = CreateMatrix.DenseOfColumnMajor(yData.Count(), 1, yData);

        Matrix<double> pinvMatA = matA.PseudoInverse();
        Console.WriteLine(pinvMatA);

        Matrix<double> matX = pinvMatA * matB;
        return matX.AsColumnMajorArray();
    }

    private static void DrawPlotChart(double[] xData, double[] yData, Func<double, double>[] drawFunc)
    {
        string chartFileName = "click.svg";

        int xMin = 0;
        int yMin = 0;

        int xMax = (int)xData.Max() + 10;
        int yMax = (int)yData.Max() + 10;

        int colorIdx = 0;
        int[] lineColors = new int[] { PLplot.Color.Blue, PLplot.Color.Red, PLplot.Color.Yellow };

        using (var pl = new PLStream())
        {
            pl.sdev("svg");
            pl.sfnam(chartFileName);

            pl.spal0("cmap0_alternate.pal");
            pl.init();

            pl.env(xMin, xMax, yMin, yMax, AxesScale.Independent, AxisBox.BoxTicksLabelsAxes);

            pl.lab("X", "Y", "Click");

            char code = Symbol.Bullet;
            pl.spal0("");
            pl.col0(lineColors[colorIdx]);
            pl.poin(xData, yData, code);

            {
                double[] ptX = PyhtonUtils.RangePython(xMin, xMax, 0.1).ToArray();
                double[] ptY = null;

                foreach (var func in drawFunc)
                {
                    pl.col0(lineColors[colorIdx % (lineColors.Length)]);
                    ptY = new double[ptX.Length];
                    for (int i = 0; i < ptX.Length; i++)
                    {
                        ptY[i] = func(ptX[i]);
                    }

                    pl.line(ptX, ptY);
                    colorIdx++;
                }

            }

            pl.eop();
            pl.gver(out var verText);
        }
    }
}
